#! /usr/bin/env node
/*
 * Copyright 2017. Tooltwist, Ltd.
 *
 * Licensed under the MIT License (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://opensource.org/licenses/MIT
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

//
//  If you wish to understand this code, this link will be invaluable:
//  http://docs.aws.amazon.com/AWSJavaScriptSDK/latest/AWS/EC2.html#describeSubnets-property
//
"use strict";

const AWS = require('aws-sdk');
const awsRegions = require('aws-regions')
const express = require('express')
const app = express()


// Load the SDK and set the initial region
var INITIAL_REGION = 'ap-northeast-1';
var ec2 = null;
var nodeIndex = [ ]; // key -> AWSNode
var regions = [ ]
var currentRegion = INITIAL_REGION;


// AWS Node types
const SUBNET = 'Subnet'
const SECGRP = 'Security Group'
const VPC = 'Virtual Private Cloud'
const NAT = 'NAT Gateway'
const IGW = 'Internet Gateway'
const INSTANCE = 'EC2 Instance'
const IMAGE = 'AMI Image'
const AZ = 'Availability Zone'
const ADDR = 'Elastic IP'
const COSTCD = 'Cost Code'
const LOGICAL = 'Logical Grouping'
const PUBLICIP = 'Public IP Address'
const KEYPAIR = 'Key Pair'
const NETIFACE = 'Network Interface'
const ROUTETABLE = 'Route Table'
const ALB = 'Application Load Balancer'
const TARGETGRP = 'Target Group'



class AWSNode {

  constructor(type, id) {
    // console.log('AWSNode.constructor()');
    this.type = type
    this.id = id
    this.key = locator(type, id)
    this.data = null
    this.children = [ ]
    this.parents = [ ]
    this.describeFn = null
  }

  setData(data) {
    this.data = data
  }

  setDescribe(fn) {
    this.describeFn = fn
  }

  addChild(childNode) {
    if (childNode != null) {
      this.children.push(childNode)
      childNode.parents.push(this)
    }
  }

  findTag(tagname) {
    // Assumes this.data.Tags
    if (this.data && this.data.Tags) {
      for (var i = 0; i < this.data.Tags.length; i++) {
        let tag = this.data.Tags[i]
        if (tag.Key === tagname) {
          return tag.Value
        }
      }
    }
    return null
  }

  awsLink() {
    let ec2Home = 'https://' + currentRegion + '.console.aws.amazon.com/ec2/v2/home?region=' + currentRegion
    let vpcHome = 'https://' + currentRegion + '.console.aws.amazon.com/vpc/home?region=' + currentRegion
    switch (this.type) {
      case SUBNET:
        return vpcHome + '#subnets:filter=' + this.id
      case SECGRP:
        return ec2Home + '#SecurityGroups:groupId=' + this.id
      case VPC:
        return vpcHome + '#vpcs:filter=' + this.id
      case NAT:
        return vpcHome + '#NatGateways:search=' + this.id + ';sort=natGatewayId'
      case IGW:
        return vpcHome + '#igws:filter=' + this.id
      case INSTANCE:
        return ec2Home + '#Instances:search=' + this.id
      // case IMAGE = 'AMI Image'
      // case AZ = 'Availability Zone'
      case ADDR:
        return ec2Home + '#Addresses:search=' + this.id
      // case COSTCD = 'Cost Code'
      // case LOGICAL = 'Logical Grouping'
      // case PUBLICIP = 'Public IP Address'
      case KEYPAIR:
        return ec2Home + '#KeyPairs:search=' + this.id
      case NETIFACE:
        return ec2Home + '#NIC:networkInterfaceId=' + this.id
      case ROUTETABLE:
        return vpcHome + '#routetables:filter=' + this.id
      default:
        return null
    }
  }

  dump() {
    if (this.describeFn) {
      let desc = this.describeFn(this)
      if (desc) {
        console.log(desc);
      }
    } else {
      console.log('Id: ' + this.id)
    }
    // console.log(' - children:', this.children)
    this.children.forEach(child => {
      // console.log('  ->', child)
      console.log('  - ' + child.id)
    })
  }
}//- class AWSNode

function locator(type, id) {
  return type + '::' + id
}

function findNode(type, id, data, describeFn) {
  let key = locator(type, id)
  let node = nodeIndex[key]
  if (!node) {
    node = new AWSNode(type, id)
    nodeIndex[key] = node
  }
  if (data) {
    node.setData(data)
  }
  if (describeFn) {
    node.setDescribe(describeFn)
  }
  return node
}

function dumpAll() {
  console.log('dumpAll');
  // console.log('nodeIndex=', nodeIndex);
  for (var key in nodeIndex) {
    console.log('')
    // console.log('key: ' + key);
    let node = nodeIndex[key]
    node.dump()
  }
}

function downloadVpcs(callback) {
  console.log('  downloadVpcs()');
  ec2.describeVpcs({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.Vpcs.forEach(rec => {
      // console.log('rec=', rec);
      findNode(VPC, rec.VpcId, rec, (node) => {
        // Return a description
        let desc = 'Virtual Private Cloud (' + node.id + ')\n'
        let name = node.findTag('Name');
        if (name) {
          desc += '  Name: ' + name + '\n'
        }
        return desc
      })
    })
    return callback(null)
  })
}

function downloadSubnets(callback) {
  console.log('  downloadSubnets()');
  ec2.describeSubnets({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.Subnets.forEach(rec => {
      let sn = findNode(SUBNET, rec.SubnetId, rec, node => {
        // Return a description
        let desc = 'Subnet (' + node.id + ')';

        // See if the subnet's routing table connects to an Internet Gateway
        let isPublicSubnet = false
        node.parents.forEach(parent => {
          if (parent.type === ROUTETABLE && parent.hasPublicRoutes) {
            isPublicSubnet = true
          }
        })
        if (isPublicSubnet) {
          desc += '  --  PUBLIC SUBNET'
        }

        let name = node.findTag('Name')
        if (name) {
          desc += '\nName: ' + name
        }

        // Route table?
        let useVpcRouteTable = true
        node.parents.forEach(parent => {
          if (parent.type === ROUTETABLE) {
              useVpcRouteTable = false
          }
        })
        if (useVpcRouteTable) {
          desc += '\nUse default route table for VPC'
        }
        // IP Address range
        desc += '\nCidrBlock: ' + node.data.CidrBlock
        return desc
      })

      // Add this to it's VPC
      let v = findNode(VPC, rec.VpcId, null)
      v.addChild(sn)

      // Add this to it's AvailabilityZone
      let az = findNode(AZ, rec.AvailabilityZone, null)
      az.addChild(sn)
    })
    return callback(null)
  })
}

function downloadSecurityGroups(callback) {
  console.log('  downloadSecurityGroups()');
  ec2.describeSecurityGroups({}, (err, data) => {
    if (err) return callback(err);
    data.SecurityGroups.forEach(grp => {
      // console.log('grp=', grp);
      let g = findNode(SECGRP, grp.GroupId, grp, node => {
        // Return a description
        let desc = 'Security Group (' + node.id + ')\n'
        desc += '  Name: ' + node.data.GroupName + '\n'
        desc += '  Description: ' + node.data.Description + '\n'
        return desc
      })

      // Add this to it's VPC
      let v = findNode(VPC, grp.VpcId, null)
      v.addChild(g)
    })
    return callback(null)
  })
}

function downloadNatGateways(callback) {
  console.log('  downloadNatGateways()');
  ec2.describeNatGateways({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.NatGateways.forEach(grp => {
      let g = findNode(NAT, grp.NatGatewayId, grp, node => {
        // Return a description
        let desc = 'NAT Gateway (' + node.id + ')\n'
        desc += '  State: ' + node.data.State + '\n';
        return desc
      })

      if (grp.State !== 'deleted') {
        // Add this to it's subnet
        let sn = findNode(SUBNET, grp.SubnetId, null)
        sn.addChild(g)

        // Add this to it's VPC
        let v = findNode(VPC, grp.VpcId, null)
        v.addChild(g)
      }
    })
    return callback(null)
  })
}

function downloadInternetGateways(callback) {
  console.log('  downloadInternetGateways()');
  ec2.describeInternetGateways({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.InternetGateways.forEach(rec => {
      let igw = findNode(IGW, rec.InternetGatewayId, rec, node => {
        // Return a description
        let desc = 'Internet Gateway (' + node.id + ')\n'
        let name = node.findTag('Name')
        if (name) {
          desc += '  Name: ' + name + '\n'
        }
        return desc
      })

      // See where it is attached
      rec.Attachments.forEach(attachment => {
        if (attachment.VpcId) {
          let vpc = findNode(VPC, attachment.VpcId)
          vpc.addChild(igw)
        }
      })
    })
    return callback(null)
  })
}

function downloadInstances(callback) {
  console.log('  downloadInstances()');
  ec2.describeInstances({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.Reservations.forEach(res => {
      res.Instances.forEach(instance => {
        // console.log('instance=', instance);
        let i = findNode(INSTANCE, instance.InstanceId, instance, node => {
          // Return a description
          let desc = 'EC2 Instance (' + node.id + ')\n';
          let name = node.findTag('Name')
          if (name) {
            desc += '  Name: ' + name + '\n';
          }
          return desc
        })

        // ImageId
        let img = findNode(IMAGE, instance.ImageId, null)
        img.addChild(i)

        // Availability zone
        if (instance.Placement && instance.Placement.AvailabilityZone) {
          let az = findNode(AZ, instance.Placement.AvailabilityZone, null, node => {
            // Return a description
            let desc = 'Availability Zone (' + node.id + ')\n'
            return desc
          })
          az.addChild(i)
        }

        // Remember if it has a public IP
        if (instance.PublicIpAddress) {
          let ip = findNode(PUBLICIP, instance.PublicIpAddress)
          ip.addChild(i)
        }

        // Subnet
        //ZZZ Can an instance have multiple subnets?
        let sn = findNode(SUBNET, instance.SubnetId)
        sn.addChild(i)

        // VPC
        let vpc = findNode(VPC, instance.VpcId)
        vpc.addChild(i)

        // Security groups
        instance.SecurityGroups.forEach(sg => {
          let g = findNode(SECGRP, sg.GroupId)
          g.addChild(i)
          i.addChild(g)
        })
      })
    })
    return callback(null)
  })
}

// Elastic IPs
function downloadAddresses(callback) {
  console.log('  downloadAddresses()');
  ec2.describeAddresses({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.Addresses.forEach(rec => {
      let addr = findNode(ADDR, rec.AllocationId, rec, node => {
        // Return a description
        let desc = 'Elastic IP (' + node.id + ')\n'
        desc += '  IP Address: ' + node.data.PublicIp + '\n';
        return desc
      })

      // Remember if it has a public IP
      if (rec.PublicIp) {
        let ip = findNode(PUBLICIP, rec.PublicIp)
        ip.addChild(addr)
      }
    })
    return callback(null)
  })
}

function downloadAvailabilityZones(callback) {
  console.log('  downloadAvailabilityZones()');
  ec2.describeAvailabilityZones({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.AvailabilityZones.forEach(rec => {
      let g = findNode(AZ, rec.ZoneName, rec, node => {
        // Return a description
        let desc = 'Availability Zone (' + node.id + ')\n'
        // desc += '  IP Address: ' + node.data.PublicIp + '\n';
        return desc
      })
    })
    return callback(null)
  })
}

function downloadKeyPairs(callback) {
  console.log('  downloadKeyPairs()');
  ec2.describeKeyPairs({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.KeyPairs.forEach(rec => {
      let g = findNode(KEYPAIR, rec.KeyName, rec, node => {
        // Return a description
        let desc = 'Key Pair (' + node.id + ')\n'
        return desc
      })
    })
    return callback(null)
  })
}

function downloadNetworkInterfaces(callback) {
  console.log('  downloadNetworkInterfaces()');
  ec2.describeNetworkInterfaces({}, (err, data) => {
    if (err) return callback(err);
    // console.log('data=', data);
    data.NetworkInterfaces.forEach(rec => {
      // console.log('\nNETWORK INTERFACE:', rec);
      let ni = findNode(NETIFACE, rec.NetworkInterfaceId, rec, node => {
        // Return a description
        let desc = 'Network Interface (' + node.id + ')\n'
        desc += node.data.Description
        return desc
      })

      // Public IP via association
      if (rec.Association && rec.Association.PublicIp) {
        let ip = findNode(PUBLICIP, rec.Association.PublicIp)
        ip.addChild(ni)
      }

      // AvailabilityZone
      let az = findNode(AZ, rec.AvailabilityZone)
      az.addChild(ni)

      // Subnet
      let subnet = findNode(SUBNET, rec.SubnetId)
      subnet.addChild(ni)

      // VPC
      let vpc = findNode(VPC, rec.VpcId)
      vpc.addChild(ni)
    })
    return callback(null)
  })
}

function downloadRegions(callback) {
  console.log('  downloadRegions()');
  ec2.describeRegions({}, (err, data) => {
    if (err) return callback(err);

    regions = data.Regions // Global variable
    // console.log('regions=', regions);
    regions = regions.sort((r1, r2) => {
      if (r1.RegionName < r2.RegionName) {
        return -1
      } else if (r1.RegionName > r2.RegionName) {
        return +1
      }
      return 0
    })
    return callback(null)
  })
}

function downloadRouteTables(callback) {
  console.log('  downloadRouteTables()');
  ec2.describeRouteTables({}, (err, data) => {
    if (err) return callback(err);

    // console.log('routeTables=', data);
    data.RouteTables.forEach(rec => {
      // console.log('\nRoute table:', rec);
      let rt = findNode(ROUTETABLE, rec.RouteTableId, rec, node => {
        // Return a description
        let desc = 'Route Table (' + node.id + ')'
        if (node.hasPublicRoutes) {
          desc += '  --  HAS PUBLIC ROUTES'
        }
        let name = node.findTag('Name')
        if (name) {
          desc += name + '\n'
        }
        // See if this is a main route table
        // node.data.Associations.forEach(assoc => {
        //   if (assoc.Main) {
        //     desc += '\nMain Route table for ' + assoc.RouteTableAssociationId
        //   }
        // })

        // Routes
        let gateways = [ ]
        desc += '\n<table class="smalltable">\n'
        node.data.Routes.forEach(route => {
          if (route.GatewayId === 'local') {
            desc += '<tr><td>' + route.DestinationCidrBlock + '</td><td>  >>>&nbsp;&nbsp;&nbsp;local</td></tr>\n'
          } else {
            let key = locator(IGW, route.GatewayId)
            desc += '<tr><td>' + route.DestinationCidrBlock + '</td><td>  >>>&nbsp;&nbsp;&nbsp;<a href="?node=' + key + '">' + route.GatewayId + '</a></td></tr>\n'
          }
        })
        desc += '</table>'
        return desc
      })

      // Look for routes to Internet Gateways
      let gateways = [ ] // Only add each gateway it once
      rec.Routes.forEach(route => {
        if (route.GatewayId != 'local') {
          if (!gateways[route.GatewayId]) {
            let igw = findNode(IGW, route.GatewayId)
            igw.addChild(rt)
            // Having a route to an Internet Gateway is the definition of a "Public Subnet"
            rt.hasPublicRoutes = true
            gateways[route.GatewayId] = true
          }
        }
      })

      // Subnets mapped to this routing table
      rec.Associations.forEach(assoc => {
        if (assoc.SubnetId) {
          let sn = findNode(SUBNET, assoc.SubnetId)
          rt.addChild(sn)
        }
      })

      // VPC
      let vpc = findNode(VPC, rec.VpcId)
      vpc.addChild(rt)
    })
    return callback(null)
  })
}

// Return the name of an AWS regions
// e.g. us-east-1 -> (n-virginia)
function regionDescription(regionCode) {
  let region = awsRegions.get(regionCode)
  if (region) {
    return '(' + region.name + ')'
  }
  return ''
}

function syntax() {
  var cmd = '    node ' + process.argv[1];
  console.log();
  console.log('Syntax:');
  console.log(cmd + '                                         - show all instances');
  console.log(cmd + ' <instanceId>                            - show a single instance');
  console.log(cmd + ' <instanceId> open [<ipaddr>] [<port>]   - open a port');
  console.log(cmd + ' <instanceId> close [<ipaddr>] [<port>]  - close a port');
  console.log();
  console.log(' Default ipaddr is your current public IP address.');
  console.log(' Default port is 22.');
  process.exit(1);
}




/*
 *  Start here.
 */
var options = { };

for (var i = 2; i < process.argv.length; i++) {
  var arg = process.argv[i];

  if (arg == '--help') {
    syntax();
  }
  else if (arg == '--dryrun') {
    options.dryrun = true;
  }
  else if (arg == 'open' || arg == 'close') {
    options.operation = arg;
  }
  else if (arg.startsWith('i-')) {
    options.instanceId = arg;
  }
  else if (!isNaN(parseInt(arg.substr(0, 1), 10))) {
    if (arg.indexOf('.') > 0) {
      // Probably an IP address
      options.myIp = arg;
    } else {
      // Must be the port
      options.port = parseInt(arg);
    }
  }
  else {
    syntax();
  }
}
// console.log('     Options: ', options);

if (options.operation) {
  if (!options.port) {
    options.port = 22;
  }
  if (!options.instanceId) {
    console.log('\nError: Missing instanceId.');
    syntax();
  }
}

// See http://docs.aws.amazon.com/AWSJavaScriptSDK/latest/AWS/EC2.html#describeSubnets-property
function downloadEverything(region, callback /*(err)*/ ) {
  console.log('Region is ' + currentRegion + ' ' + regionDescription(currentRegion));

  currentRegion = region
  ec2 = new AWS.EC2({
    apiVersion: '2016-11-15',
    region: currentRegion
  });


  nodeIndex = [ ] // Clear existing list
  downloadSecurityGroups(err => {
    if (err) return callback(err)
    downloadVpcs(err => {
      if (err) return callback(err)
      downloadSubnets(err => {
        if (err) return callback(err)
        downloadNatGateways(err => {
          if (err) return callback(err)
          downloadInternetGateways(err => {
            if (err) return callback(err)
            downloadInstances(err => {
              if (err) return callback(err)
              downloadAddresses(err => {
                if (err) return callback(err)
                downloadAvailabilityZones(err => {
                  if (err) return callback(err)
                  downloadKeyPairs(err => {
                    if (err) return callback(err)
                    downloadNetworkInterfaces(err => {
                      if (err) return callback(err)
                      downloadRegions(err => {
                        if (err) return callback(err)
                        downloadRouteTables(err => {
                          if (err) return callback(err)
                          // downloadRegions(err => {
                          //   if (err) return callback(err)

                // describeIamInstanceProfileAssociations
                // describeNetworkAcls
                // describeRouteTables
                // describeTags ???
                // describeVpnGateways


                          console.log('finished downloading');
                          return callback(null)
                        })
                      })
                    })
                  })
                })
              })
            })
          })
        })
      })
    })
  })
}

function nodesByType(type) {
  // console.log('nodesByType(' + type + ')');
  var arr = [ ]
  for (var key in nodeIndex) {
    let node = nodeIndex[key]
    if (node.type === type) {
      arr.push(node)
    }
  }
  return sortAlphabetic(arr)
}

function sortAlphabetic(arr) {
  let sorted = arr.sort((a,b) => {
    if (a.id < b.id) {
      return -1
    } else if (a.id > b.id) {
      return 1
    }
    return 0
  })
  return sorted
}

function htmlHeader(extraStyles, str) {

  // A few styles
  var html = ''
  html += '<html lang="en">\n';
  html += '<head>\n'
  html += '<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />\n'
  // See https://stackoverflow.com/questions/2467454/why-does-chrome-incorrectly-determine-page-is-in-a-different-language-and-offer
  html += '<meta charset="UTF-8">'
  html += '<meta name="google" content="notranslate">'
  html += '<meta http-equiv="Content-Language" content="en">'
  html += '<html lang="en" xml:lang="en" xmlns= "http://www.w3.org/1999/xhtml">'

  html += '</head>\n'
  html += '\n'
  html += '<style>\n';
  html += 'body {\n'
  html += '  font-family: arial narrow,helvetica;\n'
  html += '  font-size: 16px;\n'
  html += '  margin: 0px;\n'
  html += '}\n'
  html += '.header {\n'
  html += '  padding: 0px;\n'
  html += '  padding-top: 10px;\n'
  html += '  margin-botom: 0px;\n'
  html += '  backround-color: #ffffff;\n'
  html += '}\n'
  html += 'a:visited {\n'
  html += '  text-decoration: none;\n'
  html += '  color: yellow;\n'
  html += '}\n'
  html += 'a:-webkit-any-link {\n'
  html += '  text-decoration: none;\n'
  html += '  color: #333;\n'
  html += '}\n'
  html += 'h1 {\n'
  html += '  text-align: center;\n'
  html += '}\n'
  html += 'td {\n'
  html += '  vertical-align: top;\n'
  html += '}\n'
  html += '.card {\n'
  html += '  border: solid 1px #999;\n'
  html += '  padding-top: 0px;\n'
  html += '  padding-left: 10px;\n'
  html += '  padding-right: 10px;\n'
  html += '  padding-bottom: 0px;\n'
  html += '  margin: 5px;\n'
  html += '  overflow-y: scroll;\n'
  html += '}\n'
  html += '.awslink {\n'
  html += '  text-align: right;\n'
  html += '  font-size: 15px;\n'
  html += '  padding-right: 10px;\n'
  html += '}\n'
  html += '.smalltable {\n'
  html += '  font-size: 13px;\n'
  html += '}\n'
  html += '.region-select {\n'
  html += '  float: right;\n'
  html += '  margin-right: 30px;\n'
  html += '}\n'
  html += '.region-select a {\n'
  html += '  font-size: 11px;\n'
  html += '}\n'
  if (extraStyles) {
    html += extraStyles
  }
  html += '</style>';
  html += '<body id="body">';


  // Heading for the page
  let label = 'AWS Explorer'
  if (str) {
    label += ' - ' + str
  }
  html += '<div class="header">'
  html += '<h1>' + label + '</h1>\n'
  html += '<hr/>\n'
  html += '&nbsp;&nbsp;<a href="/">Index</a> &nbsp;&nbsp;&nbsp;'
  html += '<a href="?mode=vpc">Networks</a> &nbsp;&nbsp;&nbsp;'
  html += '<a href="?mode=help">Help</a> &nbsp;&nbsp;&nbsp;'

  html += '<dev class="region-select">'
  html += '<a href="?reload=true">reload</a>&nbsp;&nbsp;&nbsp;'
  html += '<select onchange="window.location = \'?region=\' + this.value; document.getElementById(\'body\').innerHTML = \'Please wait...\';">\n'
  regions.forEach(region => {
    html += '<option value="' + region.RegionName + '"';
    if (region.RegionName === currentRegion) {
      html += ' selected'
    }
    html += '>' + region.RegionName + ' ' + regionDescription(region.RegionName) + '</option>\n'
    // Endpoint
  })
  html += '</select>\n'
  html += '</dev>' //- region-select

  html += '<hr/>\n'
  html += '</div>' //header
  return html
}//- htmlHeader

function htmlFooter() {
  let html = ''
  html += '</body>\n'
  html += '</html>\n'
  return html
}

function htmlForCard(node, extraClass) {
  let html = ''
  html += '<div class="card'
  if (extraClass) {
    html += ' ' + extraClass
  }
  html += '">\n'
  html += '<a href="?node=' + node.key + '"><pre>'
  if (node.describeFn) {
    html += node.describeFn(node)
  } else {
    html += node.type + ' ' + node.id
  }
  html += '</pre></a>'
  html += '</div>\n' //- card
  return html
}

// Display a specific node
function htmlForNode(node) {
  if (!node) {
    return htmlForIndex()
  }

  let extraStyles = ''
  extraStyles += 'table {\n'
  extraStyles += '  zwidth: 100%;\n'
  extraStyles += '  zmax-width: 1280;\n'
  extraStyles += '}\n'
  extraStyles += '.pane {\n'
  extraStyles += '  padding: 20px;\n'
  extraStyles += '}\n'
  extraStyles += '.data {\n'
  extraStyles += '  max-width: 400px;\n'
  extraStyles += '  padding-top: 0px;\n'
  extraStyles += '  padding-left: 10px;\n'
  extraStyles += '  padding-right: 10px;\n'
  extraStyles += '  padding-bottom: 0px;\n'
  extraStyles += '  color: #aad;\n'
  extraStyles += '  overflow-y: scroll;\n'
  extraStyles += '}\n'
  extraStyles += '.card {\n'
  extraStyles += '  background-color: #f0f0f0;\n'
  extraStyles += '  min-width: 400px;\n'
  extraStyles += '  box-shadow: 5px 5px 2px #888888;'
  extraStyles += '  margin-bottom: 10px;\n'
  extraStyles += '}\n'
  extraStyles += '.label {\n'
  extraStyles += '  text-align: center;\n'
  extraStyles += '  min-width: 400px;\n'
  extraStyles += '}\n'
  extraStyles += '.label h3 {\n'
  extraStyles += '  color: red !important;\n'
  extraStyles += '}\n'
  extraStyles += '.card.middle {\n'
  extraStyles += '  background-color: #eef;\n'
  extraStyles += '  margin-bottom: 10px;\n'
  extraStyles += '}\n'
  extraStyles += 'pre {\n'
  extraStyles += '  font-family: arial narrow,helvetica;\n'
  extraStyles += '  font-size: 15px;\n'
  extraStyles += '}\n'


  let html = htmlHeader(extraStyles, node.type + ' ' + node.id)

  html += '<div class="pane">\n'
  html += '<table>\n'
  html += '<tr>\n'



  // Left column shows parents
  html += '<td>\n'
  html += '<div class="label"><h3>Referenced by</h3></div>'
  node.parents.forEach(parent => {
    html += htmlForCard(parent)
  })
  html += '</td>\n'  // End Left column



  // Middle column (this node)
  html += '<td>\n'
  html += '<div class="label"><h3>' + node.id + '</h3></div>\n'
  html += htmlForCard(node, "middle")
  let link = node.awsLink()
  if (link) {
    html += '<div class="awslink center"><a href="' + link + '" target="_blank">View on AWS console</a></div>\n'
  }
  if (node.data) {
    html += '<pre class="data">' + JSON.stringify(node.data, null, "  ") + '</pre>'
  }
  html += '</td>\n'  // End middle column



  // Right column shows children
  html += '<td>\n'  // Right column (children)
  html += '<div class="label"><h3>References</h3></div>'
  node.children.forEach(child => {
    html += htmlForCard(child)
  })
  html += '</td>\n'  // End right column
  html += '</tr>\n'
  html += '</table>\n'
  html += '</div>' //- pane

  // Add the end
  html += htmlFooter()
  return html
}

function htmlForNetwork() {
  // console.log('htmlForNetwork()');

  let extraStyles = ''
  extraStyles += '.pane {\n'
  extraStyles += '  background-color: white;\n'
  extraStyles += '}\n'
  extraStyles += '.card {\n'
  extraStyles += '  background-color: #f7f7f7;'
  extraStyles += '  box-shadow: 10px 10px 5px #888888;'
  extraStyles += '  margin-bottom: 30px;'
  extraStyles += '  margin-left: 50px;'
  extraStyles += '  margin-right: 70px;'
  extraStyles += '  padding-top: 15px;'
  extraStyles += '  padding-bottom: 15px;'
  extraStyles += '}\n'
  extraStyles += '.card-az {\n'
  extraStyles += '  display: inline-block;\n'
  extraStyles += '  background-color: white;\n'
  extraStyles += '  min-height: 70px;\n'
  extraStyles += '  border: dashed 1px #666;\n'
  extraStyles += '  padding-top: 10px;\n'
  extraStyles += '  padding-left: 10px;\n'
  extraStyles += '  padding-right: 10px;\n'
  extraStyles += '  padding-bottom: 8px;\n'
  extraStyles += '  margin: 15px;\n'
  extraStyles += '  overflow-y: scroll;\n'
  // extraStyles += '  box-shadow: 10px 10px 5px #888888;'
  extraStyles += '}\n'
  extraStyles += '.card-subnet {\n'
  extraStyles += '  display: inline-block;\n'
  extraStyles += '  min-height: 70px;\n'
  extraStyles += '  min-width: 160px;\n'
  extraStyles += '  border: solid 1px blue;\n'
  extraStyles += '  background-color: #eef;\n'
  extraStyles += '  padding-top: 8px;\n'
  extraStyles += '  padding-left: 10px;\n'
  extraStyles += '  padding-right: 10px;\n'
  extraStyles += '  padding-bottom: 8px;\n'
  extraStyles += '  margin: 5px;\n'
  extraStyles += '  overflow-y: scroll;\n'
  extraStyles += '}\n'

  // Get the HTML for the various types
  let html = htmlHeader(extraStyles, 'Network Topology')
  html += '<div class="pane">'
  html += '<br>'
  nodesByType(VPC).forEach(vpc => {
    html += '<div class="card">'
    html += '<a href="?node=' + vpc.key + '"><b>Virtual Private Cloud (' + vpc.id + ')</b></a><br>'
    let name = vpc.findTag('Name');
    if (name) {
      html += '  Name: ' + name + '<br>'
    }

    // Show each of the availability Zones
    nodesByType(AZ).forEach(az => {
      html += '<div class="card-az">'
      html += '<a href="?node=' + az.key + '">Availability Zone ' + az.id + '</a><br>'

      // Display the subnets in this availability zone
      // console.log('az=' + az.key);
      vpc.children.forEach(child => {
        if (child.type === SUBNET) {
            // console.log('   subnet=' + child.key);

          // See if the subnet is in this availability zone
          let subnet = child
          let inZone = false;
          subnet.parents.forEach(parent => {
            if (parent === az) {
              inZone = true
            }
          })
          if (inZone) {

            // The subnet is public if it has a public route, via a
            // routeTable that routes to an Internet Gateway.
            let isPublicSubnet = false
            subnet.parents.forEach(parent => {
              if (parent.type === ROUTETABLE && parent.hasPublicRoutes) {
                isPublicSubnet = true
              }
            })

            // Show the subnet
            html += '<div class="card-subnet">'
            html += '<a href="?node=' + subnet.key + '"><b>' + subnet.id
            html += isPublicSubnet ? '  (public)' : '  (private)'
            html += '</b></a><br>'

            // Show it's parents children
            let useVpcRouteTable = true
            // let vpcRouteTable = null
            subnet.parents.forEach(node => {
              if (node.type === VPC) {
                // console.log('VPC route table:', vpc.data);
                // vpcRouteTable = 'xyz'
              }
              else if (node.type === AZ) {
                // Ignore
              } else {
                html += '  <a href="?node=' + node.key + '">' + node.type + ' ' + node.id + '</a><br>'
                if (node.type === ROUTETABLE) {
                  useVpcRouteTable = false
                }
              }
            })
            if (useVpcRouteTable) {
              html += 'Use default route table for VPC<br>'
            }
            subnet.children.forEach(node => {
              html += '  <a href="?node=' + node.key + '">' + node.type + ' ' + node.id + '</a><br>'
            })
            html += '</div>' // card for subnet
          }
        }
      })
      html += '</div>' //- card for AZ
    })
    html += '</div>' // card for VPC
    html += '<br>'

  })
  html += '</div>' //- pane
  html += htmlFooter()
  return html
}

function htmlForIndex() {

  // Get the HTML for the various types
  let known = [ ]
  let instance = '<h3>Instance</h3>'
  nodesByType(INSTANCE).forEach(node => {
    instance += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let nat = '<h3>NAT Gateways</h3>'
  nodesByType(NAT).forEach(node => {
    nat += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let igw = '<h3>Internet Gateways</h3>'
  nodesByType(IGW).forEach(node => {
    igw += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let image = '<h3>AMI Images</h3>'
  nodesByType(IMAGE).forEach(node => {
    image += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let vpc = '<h3>VPCs</h3>'
  nodesByType(VPC).forEach(node => {
    vpc += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let az = '<h3>Availability Zones</h3>'
  nodesByType(AZ).forEach(node => {
    az += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let secgrp = '<h3>Security Groups</h3>'
  nodesByType(SECGRP).forEach(node => {
    secgrp += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let subnet = '<h3>Subnets</h3>'
  nodesByType(SUBNET).forEach(node => {
    subnet += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let addr = '<h3>Elastic IPs</h3>'
  nodesByType(ADDR).forEach(node => {
    addr += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let publicip = '<h3>Public IP Addresses</h3>'
  nodesByType(PUBLICIP).forEach(node => {
    publicip += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let keypair = '<h3>Key Pairs</h3>'
  nodesByType(KEYPAIR).forEach(node => {
    keypair += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let netiface = '<h3>Network Interfaces</h3>'
  nodesByType(NETIFACE).forEach(node => {
    netiface += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })
  let routeTables = '<h3>Route Tables</h3>'
  nodesByType(ROUTETABLE).forEach(node => {
    routeTables += '<a href="?node=' + node.key + '">' + node.id + '</a><br>'
    known[node.key] = true
  })


  // Display all nodes
  var extraStyles = ''
  extraStyles += 'body {\n'
  extraStyles += '  font-family: arial narrow,helvetica;\n'
  extraStyles += '  font-size: 16px;\n'
  extraStyles += '}\n'
  extraStyles += 'td {\n'
  extraStyles += '  vertical-align: top;\n'
  extraStyles += '  padding-right: 80px;\n'
  extraStyles += '}\n'
  extraStyles += '.pane {\n'
  extraStyles += '  padding: 20px'
  extraStyles += '}\n'

  let currentNode = null
  let html = htmlHeader(extraStyles, currentNode)
  html += '<div class="pane">'
  html += '<table>'
  html += '<tr>'

  html += '<td>'
  html += vpc
  html += az
  html += subnet
  html += '</td>'

  html += '<td>'
  html += secgrp
  html += igw
  html += nat
  html += '</td>'

  html += '<td>'
  html += addr
  html += publicip
  html += netiface
  html += routeTables
  html += '</td>'

  // Display anything else
  html += '<td>'
  html += instance
  html += image
  html += keypair
  html += '<br>'
  html += '<br>'
  html += '<h3>Other</h3>';
  for (var key in nodeIndex) {
    // console.log('')
    // console.log('key: ' + key);
    // let node = nodeIndex[key]
    // node.dump()
    if (!known[key]) {
      let node = nodeIndex[key]
      let numParents = node.parents.length
      let numChildren = node.children.length
      html += '<a href="?node=' + key + '">' + node.id + '</a> [' + numParents + ',' + numChildren + ']<br>'
    }
  }
  html += '</td>'

  html += '</tr>'
  html += '</table>'
  html += '</div>' //- pane

  // Add the end
  html += htmlFooter()
  return html
} //- htmlForIndex


function htmlForHelp() {

  let extraStyles = ''
  extraStyles += '.pane a:hover {\n'
  extraStyles += '  text-decoration: underline;\n'
  extraStyles += '  color: blue;\n'
  extraStyles += '}\n'
  extraStyles += '.pane {\n'
  extraStyles += '  padding: 20px;\n'
  extraStyles += '  padding-left: 120px;\n'
  extraStyles += '  font-size: 18px;\n'
  extraStyles += '}\n'
  extraStyles += 'h3 {\n'
  extraStyles += '  margin-bottom: 15px;\n'
  extraStyles += '}\n'

  let heading = null
  let html = htmlHeader(extraStyles, heading)
  html += '<div class="pane">'
  html += '<h2>Help Topics</h2>'
  html += '<h3>Overview of AWS</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://aws.amazon.com/ecs/getting-started/" target="_blank">'
  html += 'Getting Started</a><br>\n'

  html += '<h3>Security</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/aws-security-groups-instance-level-security/" target="_blank">'
  html += 'Instance level security (Security groups)</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/aws-network-acl-vpc-subnets-network-security/" target="_blank">'
  html += 'Network level security (Routing tables and Network ACLs)</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/how-to-build-hipaa-compliant-applications-on-aws/" target="_blank">'
  html += 'How to build HIPAA compliant applications on AWS</a><br>\n'

  html += '<h3>Load Balancing</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/application-load-balancer-vs-classic-load-balancer/" target="_blank">'
  html += 'Application Load Balancer vs Classic Load Balancer</a><br>\n'

  html += '<h3>Containers</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/amazon-ec2-container-service-docker-aws/" target="_blank">'
  html += 'Amazon EC2 Container Service and Elastic Beanstalk: Docker on AWS</a><br>\n'

  html += '<h3>Cloudformation</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://cloudacademy.com/blog/cloudformation-deployment-tool/" target="_blank">'
  html += 'CloudFormation: the perfect Cloud Deployment tool for AWS?</a><br>\n'

  html += '<h3>Deployment</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://aws.amazon.com/blogs/compute/continuous-deployment-to-amazon-ecs-using-aws-codepipeline-aws-codebuild-amazon-ecr-and-aws-cloudformation/" target="_blank">'
  html += 'Continuous Deployment to Amazon ECS using AWS CodePipeline, AWS CodeBuild, Amazon ECR, and AWS CloudFormation</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://github.com/awslabs/ecs-refarch-continuous-deployment" target="_blank">'
  html += 'ECS Reference Architecture: Continuous Deployment</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://github.com/awslabs/golang-deployment-pipeline" target="_blank">'
  html += 'Deployment Pipeline for Go Applications on AWS</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="http://docs.aws.amazon.com/devops/latest/gsg/welcome.html" target="_blank">'
  html += 'Introducing the AWS for DevOps Walkthrough</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="http://docs.aws.amazon.com/devops/latest/gsg/welcome.html" target="_blank">'
  html += 'Introducing the AWS for DevOps Walkthrough</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="http://docs.aws.amazon.com/devops/latest/gsg/welcome.html" target="_blank">'
  html += 'Introducing the AWS for DevOps Walkthrough</a><br>\n'

  html += '<h3>Tooltwist</h3>'
  html += '   &nbsp;&nbsp;&nbsp;<a href="http://tooltwist.com" target="_blank">'
  html += 'Tooltwist website</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://github.com/tooltwist/aws-explorer/wiki" target="_blank">'
  html += 'aws-explorer wiki</a><br>\n'
  html += '   &nbsp;&nbsp;&nbsp;<a href="https://github.com/tooltwist/aws-explorer/wiki/Tooltwist-Cloudformation-Templates" target="_blank">'
  html += 'Tooltwist Cloudformation Templates</a><br>\n'



  html += '</div>' //- pane

  // Add the end
  html += htmlFooter()
  return html
} //- htmlForHelp



//
//
//    Here is the start of our code.
//
//
console.log();
console.log();
console.log('Please wait for the initial download...');
console.log();
downloadEverything(INITIAL_REGION, err => {
  if (err) {
    console.log('Error:', err);
  } else {
    // dumpAll()

    app.get('/', function (req, res) {

      // See if we're changing region
      if (req.query.reload) {

        /*
         *  Reload from AWS
         */
        downloadEverything(currentRegion, err => {
          if (err) {
            console.log('Error:', err);
          } else {
            res.send(htmlForIndex())
          }
        })
      } else if (req.query.region && req.query.region !== currentRegion) {

        /*
         *  Load a new Region
         */
        downloadEverything(req.query.region, err => {
          if (err) {
            console.log('Error:', err);
          } else {
            res.send(htmlForIndex())
          }
        })
      } else if (req.query.mode && req.query.mode === 'help') {

        /*
         *  Display the help page.
         */
        res.send(htmlForHelp())
      } else if (req.query.mode && req.query.mode === 'vpc') {

        /*
         *  Display the network hierarchy.
         */
        res.send(htmlForNetwork())
      } else if (req.query.node){

        /*
         *  Displaying a specific node
         */
        console.log('Selected node ' + req.query.node);
        let node = nodeIndex[req.query.node]
        res.send(htmlForNode(node))
      } else {

        /*
         *  Show the index page
         */
        res.send(htmlForIndex())
      }
    })

    app.listen(3000, function () {
      console.log();
      console.log();
      console.log('Listening now on port 3000...')
    })

  }
})
